from __future__ import division

from face_detector.models import *
from face_detector.utils.utils import *
from face_detector.utils.datasets import *
from face_detector.utils.parse_config import *

import os
import sys
import time
import datetime
import argparse
import tqdm

import torch
from torch.utils.data import DataLoader
from torch.autograd import Variable
import torch.optim as optim


def evaluate(model, path, iou_thres, conf_thres, nms_thres, img_size, batch_size):
    model.eval()

    # Get dataloader
    dataset = ListDataset(path, img_size=img_size, augment=False, multiscale=False)
    dataloader = torch.utils.data.DataLoader(
        dataset, batch_size=batch_size, shuffle=False, num_workers=1, collate_fn=dataset.collate_fn
    )

    Tensor = torch.cuda.FloatTensor if torch.cuda.is_available() else torch.FloatTensor

    labels = []
    sample_metrics = []  # List of tuples (TP, confs, pred)
    for batch_i, (_, imgs, targets) in enumerate(tqdm.tqdm(dataloader, desc="Detecting objects")):

        # Extract labels
        labels += targets[:, 1].tolist()
        # Rescale target
        targets[:, 2:] = xywh2xyxy(targets[:, 2:])
        targets[:, 2:] *= img_size

        imgs = Variable(imgs.type(Tensor), requires_grad=False)

        with torch.no_grad():
            outputs = model(imgs)
            outputs = non_max_suppression(outputs, conf_thres=conf_thres, nms_thres=nms_thres)

        sample_metrics += get_batch_statistics(outputs, targets, iou_threshold=iou_thres)

    # Concatenate sample statistics
    true_positives, pred_scores, pred_labels = [np.concatenate(x, 0) for x in list(zip(*sample_metrics))]
    precision, recall, AP, f1, ap_class = ap_per_class(true_positives, pred_scores, pred_labels, labels)

    return precision, recall, AP, f1, ap_class


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--batch_size", type=int, default=8, help="size of each image batch")
    parser.add_argument("--model_def", type=str, default="config/yolov3.cfg", help="path to model definition file")
    parser.add_argument("--data_config", type=str, default="config/coco.data", help="path to data config file")
    parser.add_argument("--weights_path", type=str, default="weights/yolov3.weights", help="path to weights file")
    parser.add_argument("--class_path", type=str, default="data/coco.names", help="path to class label file")
    parser.add_argument("--iou_thres", type=float, default=0.5, help="iou threshold required to qualify as detected")
    parser.add_argument("--conf_thres", type=float, default=0.001, help="object confidence threshold")
    parser.add_argument("--nms_thres", type=float, default=0.5, help="iou thresshold for non-maximum suppression")
    parser.add_argument("--n_cpu", type=int, default=8, help="number of cpu threads to use during batch generation")
    parser.add_argument("--img_size", type=int, default=416, help="size of each image dimension")
    opt = parser.parse_args()
    print(opt)

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    data_config = parse_data_config(opt.data_config)
    valid_path = data_config["valid"]
    class_names = load_classes(data_config["names"])

    # Initiate model
    model = Darknet(opt.model_def).to(device)
    if opt.weights_path.endswith(".weights"):
        # Load darknet weights
        model.load_darknet_weights(opt.weights_path)
    else:
        # Load checkpoint weights
        model.load_state_dict(torch.load(opt.weights_path))

    print("Compute mAP...")

    precision, recall, AP, f1, ap_class = evaluate(
        model,
        path=valid_path,
        iou_thres=opt.iou_thres,
        conf_thres=opt.conf_thres,
        nms_thres=opt.nms_thres,
        img_size=opt.img_size,
        batch_size=8,
    )

    print("Average Precisions:")
    for i, c in enumerate(ap_class):
        print(f"+ Class '{c}' ({class_names[c]}) - AP: {AP[i]}")

    print(f"mAP: {AP.mean()}")



import numpy as np
import torch
from torch import nn
import torch.nn.functional as F

class EyeClassifierModel(torch.nn.Module):
  """A simple convolutional network.
  
  Map from inputs with shape [batch_size, 1, height, width] to
  outputs with shape [batch_size, 1].
  """
  
  def __init__(self):
    super().__init__()
    self.conv1=nn.Conv2d(3,96,5,3,2)
    self.conv2=nn.Conv2d(96,192,5,padding=2)
    self.conv3=nn.Conv2d(192,256,3,padding=1)
    self.conv4=nn.Conv2d(256,256,3,padding=1)
    self.conv5=nn.Conv2d(256,128,3,padding=1)
    self.fc = nn.Sequential(
            nn.Dropout(p=0.1),
            nn.Linear(8192, 4096),
            nn.ReLU(inplace=True),
            nn.Dropout(p=0.1),
            nn.Linear(4096, 4096),
            nn.ReLU(inplace=True),
            nn.Linear(4096, 2),
            nn.LogSoftmax(dim=1)
    )
    
  def forward(self, x):
    out = self.conv1(x)
    out = F.relu(out)
    out = F.max_pool2d(out, 3, 2)
    out = self.conv2(out)
    out = F.relu(out)
    out = F.max_pool2d(out, 3, 2)
    out = self.conv3(out)
    out = F.relu(out)
    out = self.conv4(out)
    out = F.relu(out)
    out = self.conv5(out)
    out = F.relu(out)
    out = F.max_pool2d(out, 3, 2)
    out = torch.flatten(out, 1)
    out = self.fc(out)
    return out



from eyeClassifierModel import EyeClassifierModel
import numpy as np
import torch
from torch import nn
from torch import optim
from torchvision import datasets, transforms, models
from torch.utils.data import Dataset, DataLoader
from torch.utils.data.sampler import SubsetRandomSampler

def load_split_train_test(datadir, trans, test_size = .2, batch=64):
    data = datasets.ImageFolder(datadir,transform=trans)
    num_data = len(data)
    print(num_data)
    indices = list(range(num_data))
    split = int(np.floor(test_size * num_data))
    np.random.shuffle(indices)
    train_idx, test_idx = indices[split:], indices[:split]
    train_sampler = SubsetRandomSampler(train_idx)
    test_sampler = SubsetRandomSampler(test_idx)
    trainloader = torch.utils.data.DataLoader(data, sampler=train_sampler, batch_size=batch, num_workers=4)
    testloader = torch.utils.data.DataLoader(data, sampler=test_sampler, batch_size=batch, num_workers=4)
    return trainloader, testloader

# Data Normalization
def data_normalizer():
    return transforms.Compose([transforms.Resize(255),
        transforms.CenterCrop(224),
        transforms.ToTensor(),
        transforms.Normalize((0.485, 0.456, 0.406), (0.229, 0.224, 0.225)),])

if __name__ == '__main__':
    BATCH_SIZE=64
    data_dir = "./data/classified_data/"

    trans = data_normalizer()
    train_loader, val_loader = load_split_train_test(data_dir, trans, .2, BATCH_SIZE)

    model = EyeClassifierModel()
    model.load_state_dict(torch.load('eye_classifier.pth'))

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    model.to(device)

    criterion = nn.NLLLoss()
    optimizer = optim.Adam(model.parameters(), lr=0.001)

    epochs = 5
    for epoch in range(epochs):
        train_loss = 0
        val_loss = 0
        accuracy = 0

        model.train()
        counter = 0
        for inputs, labels in train_loader:
            # Move to device
            inputs, labels = inputs.to(device), labels.to(device)
            # Clear optimizers
            optimizer.zero_grad()
            # Forward pass
            output = model.forward(inputs)
            # Loss
            loss = criterion(output, labels)
            # Calculate gradients (backpropogation)
            loss.backward()
            # Adjust parameters based on gradients
            optimizer.step()
            # Add the loss to the training set's rnning loss
            train_loss += loss.item()*inputs.size(0)
            
            # Print the progress of our training
            counter += 1
            if counter % 10 == 0:
                print(counter, "/", len(train_loader))
            
        # Evaluating the model
        model.eval()
        counter = 0
        # Tell torch not to calculate gradients
        with torch.no_grad():
            for inputs, labels in val_loader:
                # Move to device
                inputs, labels = inputs.to(device), labels.to(device)
                # Forward pass
                output = model.forward(inputs)
                # Calculate Loss
                valloss = criterion(output, labels)
                # Add loss to the validation set's running loss
                val_loss += valloss.item()*inputs.size(0)
                
                # Since our model outputs a LogSoftmax, find the real 
                # percentages by reversing the log function
                # output = torch.exp(output)
                # Get the top class of the output
                top_p, top_class = output.topk(1, dim=1)
                # See how many of the classes were correct?
                equals = top_class == labels.view(*top_class.shape)
                # Calculate the mean (get the accuracy for this batch)
                # and add it to the running accuracy for this epoch
                accuracy += torch.mean(equals.type(torch.FloatTensor)).item()

                
                # Print the progress of our evaluation
                counter += 1
                if (counter % 10 == 0):
                    print(counter, "/", len(val_loader))
        
        # Get the average loss for the entire epoch
        train_loss = train_loss/len(train_loader.dataset)
        valid_loss = val_loss/len(val_loader.dataset)
        # Print out the information
        print('Accuracy: ', accuracy/len(val_loader))
        print('Epoch: {} \tTraining Loss: {:.6f} \tValidation Loss: {:.6f}'.format(epoch, train_loss, valid_loss))

    # Save the model after training
    torch.save(model.state_dict(), 'eye_classifier_more_training.pth')


import cv2
import numpy as np
import matplotlib.pyplot as plt
import sys

from face_detector.interface import FaceDetectorInterface
from smile_classifier.interface import is_smiling
from eye_detector.interface import EyeDetectorInterface
from eye_classifier.interface import EyeClassifierInterface

from PIL import Image


def draw(results, image):
    font_scale = 0.5
    thickness = 1
    for face_res in results:
        cv2.rectangle(image, face_res['face_start'], face_res['face_end'], (0, 255, 0), 2)
        smiling_status = 'smiling' if face_res['smiling'] else 'not smiling'
        cv2.putText(image, smiling_status, face_res['face_start'],
                    cv2.FONT_HERSHEY_SIMPLEX, font_scale, (0, 0, 0),
                    thickness, cv2.LINE_AA)

        if face_res['eye1_start'] and face_res['eye1_end']:
            cv2.rectangle(image, face_res['eye1_start'], face_res['eye1_end'], (0, 255, 0), 2)
            eye_status = 'open' if face_res['eye1_open'] else 'closed'
            cv2.putText(image, eye_status, face_res['eye1_start'],
                        cv2.FONT_HERSHEY_SIMPLEX, font_scale, (0, 0, 0),
                        thickness, cv2.LINE_AA)

        if face_res['eye2_start'] and face_res['eye2_end']:
            cv2.rectangle(image, face_res['eye2_start'], face_res['eye2_end'], (0, 255, 0), 2)
            eye_status = 'open' if face_res['eye2_open'] else 'closed'
            cv2.putText(image, eye_status, face_res['eye2_start'],
                        cv2.FONT_HERSHEY_SIMPLEX, font_scale, (0, 0, 0),
                        thickness, cv2.LINE_AA)

    return image


def run(img=None, image_path=None, data_folder='model_data'):
    final_results = []

    if image_path:
        img = np.array(Image.open(image_path))

    gray_img = cv2.cvtColor(img, cv2.COLOR_RGB2GRAY)

    face_detector = FaceDetectorInterface.create(data_folder)
    eye_detector = EyeDetectorInterface.create(data_folder)
    eye_classifier = EyeClassifierInterface.create(data_folder)

    res = face_detector.run(img)
    for result in res:
        one_face_results = {
            'face_start': (result['x1'], result['y1']),
            'face_end': (result['x2'], result['y2']),
            'eye1_start': None,
            'eye1_end': None,
            'eye2_start': None,
            'eye2_end': None,
            'smiling': None,
            'eye1_open': None,
            'eye2_open': None
        }

        x1 = result['x1']
        x2 = result['x2']
        y1 = result['y1']
        y2 = result['y2']

        # extract the actual face image from the box coordinates
        face = cv2.getRectSubPix(gray_img,
                                 (x2 - x1, y2 - y1),
                                 ((x2 + x1) / 2, (y2 + y1) / 2))

        color_face = cv2.getRectSubPix(cv2.cvtColor(img, cv2.COLOR_RGBA2RGB),
                                        (x2 - x1, y2 - y1),
                                        ((x2 + x1) / 2, (y2 + y1) / 2))

        one_face_results['smiling'] = True if is_smiling(face, data_folder) else False

        try:
            eyes, radius = eye_detector.detect(face)
            eye1, eye2 = eyes
            diameter = 2 * radius
        except:
            continue
            

        # Drawing eyes on image
        for i, eye in enumerate([eye1, eye2]):
            start = eye[0] - radius + x1, eye[1] - radius + y1
            end = eye[0] + radius + x1, eye[1] + radius + y1
            one_face_results['eye%d_start' % (i+1)] = start
            one_face_results['eye%d_end' % (i+1)] = end

        eye1 = cv2.getRectSubPix(color_face, (diameter, diameter), tuple(eye1))
        eye2 = cv2.getRectSubPix(color_face, (diameter, diameter), tuple(eye2))

        eye1_result = eye_classifier.run(Image.fromarray(eye1))['label']
        eye2_result = eye_classifier.run(Image.fromarray(eye2))['label']

        one_face_results['eye1_open'] = True if eye1_result == 1 else False
        one_face_results['eye2_open'] = True if eye2_result == 1 else False

        final_results.append(one_face_results)

    return final_results


if __name__ == '__main__':
    image_path = sys.argv[1]
    img = np.array(Image.open(image_path))
    res = run(img)
    print(res)
    image = draw(res, img)
    plt.imshow(image)
    plt.show()



import glob
import random
import os
import sys
import numpy as np
from PIL import Image
import torch
import torch.nn.functional as F

from face_detector.utils.augmentations import horisontal_flip
from torch.utils.data import Dataset
import torchvision.transforms as transforms


def pad_to_square(img, pad_value):
    c, h, w = img.shape
    dim_diff = np.abs(h - w)
    # (upper / left) padding and (lower / right) padding
    pad1, pad2 = dim_diff // 2, dim_diff - dim_diff // 2
    # Determine padding
    pad = (0, 0, pad1, pad2) if h <= w else (pad1, pad2, 0, 0)
    # Add padding
    img = F.pad(img, pad, "constant", value=pad_value)

    return img, pad


def resize(image, size):
    image = F.interpolate(image.unsqueeze(0), size=size, mode="nearest").squeeze(0)
    return image


def random_resize(images, min_size=288, max_size=448):
    new_size = random.sample(list(range(min_size, max_size + 1, 32)), 1)[0]
    images = F.interpolate(images, size=new_size, mode="nearest")
    return images


class ImageFolder(Dataset):
    def __init__(self, folder_path, img_size=416):
        self.files = sorted(glob.glob("%s/*.*" % folder_path))
        self.img_size = img_size

    def __getitem__(self, index):
        img_path = self.files[index % len(self.files)]
        # Extract image as PyTorch tensor
        img = transforms.ToTensor()(Image.open(img_path))
        # Pad to square resolution
        img, _ = pad_to_square(img, 0)
        # Resize
        img = resize(img, self.img_size)

        return img_path, img

    def __len__(self):
        return len(self.files)


class ListDataset(Dataset):
    def __init__(self, list_path, img_size=416, augment=True, multiscale=True, normalized_labels=True):
        with open(list_path, "r") as file:
            self.img_files = file.readlines()

        self.label_files = [
            path.replace("images", "labels").replace(".png", ".txt").replace(".jpg", ".txt")
            for path in self.img_files
        ]
        self.img_size = img_size
        self.max_objects = 100
        self.augment = augment
        self.multiscale = multiscale
        self.normalized_labels = normalized_labels
        self.min_size = self.img_size - 3 * 32
        self.max_size = self.img_size + 3 * 32
        self.batch_count = 0

    def __getitem__(self, index):

        # ---------
        #  Image
        # ---------

        img_path = self.img_files[index % len(self.img_files)].rstrip()

        # Extract image as PyTorch tensor
        img = transforms.ToTensor()(Image.open(img_path).convert('RGB'))

        # Handle images with less than three channels
        if len(img.shape) != 3:
            img = img.unsqueeze(0)
            img = img.expand((3, img.shape[1:]))

        _, h, w = img.shape
        h_factor, w_factor = (h, w) if self.normalized_labels else (1, 1)
        # Pad to square resolution
        img, pad = pad_to_square(img, 0)
        _, padded_h, padded_w = img.shape

        # ---------
        #  Label
        # ---------

        label_path = self.label_files[index % len(self.img_files)].rstrip()

        targets = None
        if os.path.exists(label_path):
            boxes = torch.from_numpy(np.loadtxt(label_path).reshape(-1, 5))
            # Extract coordinates for unpadded + unscaled image
            x1 = w_factor * (boxes[:, 1] - boxes[:, 3] / 2)
            y1 = h_factor * (boxes[:, 2] - boxes[:, 4] / 2)
            x2 = w_factor * (boxes[:, 1] + boxes[:, 3] / 2)
            y2 = h_factor * (boxes[:, 2] + boxes[:, 4] / 2)
            # Adjust for added padding
            x1 += pad[0]
            y1 += pad[2]
            x2 += pad[1]
            y2 += pad[3]
            # Returns (x, y, w, h)
            boxes[:, 1] = ((x1 + x2) / 2) / padded_w
            boxes[:, 2] = ((y1 + y2) / 2) / padded_h
            boxes[:, 3] *= w_factor / padded_w
            boxes[:, 4] *= h_factor / padded_h

            targets = torch.zeros((len(boxes), 6))
            targets[:, 1:] = boxes

        # Apply augmentations
        if self.augment:
            if np.random.random() < 0.5:
                img, targets = horisontal_flip(img, targets)

        return img_path, img, targets

    def collate_fn(self, batch):
        paths, imgs, targets = list(zip(*batch))
        # Remove empty placeholder targets
        targets = [boxes for boxes in targets if boxes is not None]
        # Add sample index to targets
        for i, boxes in enumerate(targets):
            boxes[:, 0] = i
        targets = torch.cat(targets, 0)
        # Selects new image size every tenth batch
        if self.multiscale and self.batch_count % 10 == 0:
            self.img_size = random.choice(range(self.min_size, self.max_size + 1, 32))
        # Resize images to input shape
        imgs = torch.stack([resize(img, self.img_size) for img in imgs])
        self.batch_count += 1
        return paths, imgs, targets

    def __len__(self):
        return len(self.img_files)



from __future__ import division

from face_detector.models import *
from face_detector.utils.logger import *
from face_detector.utils.utils import *
from face_detector.utils.datasets import *
from face_detector.utils.parse_config import *
from face_detector.test import evaluate

from terminaltables import AsciiTable

import os
import time
import datetime
import argparse
import copy

import torch
from torch.utils.data import DataLoader
from torch.autograd import Variable


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--epochs", type=int, default=100, help="number of epochs")
    parser.add_argument("--batch_size", type=int, default=8, help="size of each image batch")
    parser.add_argument("--gradient_accumulations", type=int, default=2, help="number of gradient accums before step")
    parser.add_argument("--model_def", type=str, default="config/yolov3.cfg", help="path to model definition file")
    parser.add_argument("--data_config", type=str, default="config/coco.data", help="path to data config file")
    parser.add_argument("--pretrained_weights", type=str, help="if specified starts from checkpoint model")
    parser.add_argument("--n_cpu", type=int, default=8, help="number of cpu threads to use during batch generation")
    parser.add_argument("--img_size", type=int, default=416, help="size of each image dimension")
    parser.add_argument("--checkpoint_interval", type=int, default=1, help="interval between saving model weights")
    parser.add_argument("--evaluation_interval", type=int, default=1, help="interval evaluations on validation set")
    parser.add_argument("--compute_map", default=False, help="if True computes mAP every tenth batch")
    parser.add_argument("--multiscale_training", default=True, help="allow for multi-scale training")
    opt = parser.parse_args()
    print(opt)

    logger = Logger("logs")

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    os.makedirs("output", exist_ok=True)
    os.makedirs("checkpoints", exist_ok=True)

    # Get data configuration
    data_config = parse_data_config(opt.data_config)
    train_path = data_config["train"]
    valid_path = data_config["valid"]
    class_names = load_classes(data_config["names"])

    # Initiate model
    model = Darknet(opt.model_def).to(device)
    model.apply(weights_init_normal)

    # If specified we start from checkpoint
    if opt.pretrained_weights:
        if opt.pretrained_weights.endswith(".pth"):
            model.load_state_dict(torch.load(opt.pretrained_weights))
        else:
            model.load_darknet_weights(opt.pretrained_weights)

    # Get dataloader
    dataset = ListDataset(train_path, augment=True, multiscale=opt.multiscale_training)
    dataloader = torch.utils.data.DataLoader(
        dataset,
        batch_size=opt.batch_size,
        shuffle=True,
        num_workers=opt.n_cpu,
        pin_memory=True,
        collate_fn=dataset.collate_fn,
    )

    optimizer = torch.optim.Adam(model.parameters())

    metrics = [
        "grid_size",
        "loss",
        "x",
        "y",
        "w",
        "h",
        "conf",
        "cls",
        "cls_acc",
        "recall50",
        "recall75",
        "precision",
        "conf_obj",
        "conf_noobj",
    ]

    train_loss_list = []
    val_precision_list = []
    val_recall_list = []
    val_mAP = []
    val_f1 = []

    best_params = None
    best_metric = None

    for epoch in range(opt.epochs):
        model.train()
        start_time = time.time()
        for batch_i, (_, imgs, targets) in enumerate(dataloader):
            batches_done = len(dataloader) * epoch + batch_i

            imgs = Variable(imgs.to(device))
            targets = Variable(targets.to(device), requires_grad=False)

            loss, outputs = model(imgs, targets)
            loss.backward()

            if batches_done % opt.gradient_accumulations:
                # Accumulates gradient before each step
                optimizer.step()
                optimizer.zero_grad()

            # ----------------
            #   Log progress
            # ----------------

            log_str = "\n---- [Epoch %d/%d, Batch %d/%d] ----\n" % (epoch, opt.epochs, batch_i, len(dataloader))

            metric_table = [["Metrics", *[f"YOLO Layer {i}" for i in range(len(model.yolo_layers))]]]

            # Log metrics at each YOLO layer
            for i, metric in enumerate(metrics):
                formats = {m: "%.6f" for m in metrics}
                formats["grid_size"] = "%2d"
                formats["cls_acc"] = "%.2f%%"
                row_metrics = [formats[metric] % yolo.metrics.get(metric, 0) for yolo in model.yolo_layers]
                metric_table += [[metric, *row_metrics]]

                # Tensorboard logging
                tensorboard_log = []
                for j, yolo in enumerate(model.yolo_layers):
                    for name, metric in yolo.metrics.items():
                        if name != "grid_size":
                            tensorboard_log += [(f"{name}_{j+1}", metric)]
                tensorboard_log += [("loss", loss.item())]
                logger.list_of_scalars_summary(tensorboard_log, batches_done)

            log_str += AsciiTable(metric_table).table
            log_str += f"\nTotal loss {loss.item()}"
            train_loss_list.append(loss)

            # Determine approximate time left for epoch
            epoch_batches_left = len(dataloader) - (batch_i + 1)
            time_left = datetime.timedelta(seconds=epoch_batches_left * (time.time() - start_time) / (batch_i + 1))
            log_str += f"\n---- ETA {time_left}"

            print(log_str)

            model.seen += imgs.size(0)

        # https://github.com/eriklindernoren/PyTorch-YOLOv3/issues/222
        # Error occurs when no object is detected, which is usually the case for first epoch
        if epoch % opt.evaluation_interval == 0:
            print("\n---- Evaluating Model ----")
            # Evaluate the model on the validation set
            precision, recall, AP, f1, ap_class = evaluate(
                model,
                path=valid_path,
                iou_thres=0.5,
                conf_thres=0.5,
                nms_thres=0.5,
                img_size=opt.img_size,
                batch_size=opt.batch_size,
            )
            evaluation_metrics = [
                ("val_precision", precision.mean()),
                ("val_recall", recall.mean()),
                ("val_mAP", AP.mean()),
                ("val_f1", f1.mean()),
            ]
            logger.list_of_scalars_summary(evaluation_metrics, epoch)

            # Print class APs and mAP
            ap_table = [["Index", "Class name", "AP"]]
            for i, c in enumerate(ap_class):
                ap_table += [[c, class_names[c], "%.5f" % AP[i]]]
            print(AsciiTable(ap_table).table)
            print(f"---- mAP {AP.mean()}")

            val_precision_list.append(precision.mean())
            val_recall_list.append(recall.mean())
            val_mAP.append(AP.mean())
            val_f1.append(f1.mean())
            np.save('./WIDER/data/train_loss_list.npy', train_loss_list, allow_pickle=True)
            np.save('./WIDER/data/val_precision_list.npy', val_precision_list, allow_pickle=True)
            np.save('./WIDER/data/val_recall_list.npy', val_recall_list, allow_pickle=True)
            np.save('./WIDER/data/val_mAP.npy', val_mAP, allow_pickle=True)
            np.save('./WIDER/data/val_f1.npy', val_f1, allow_pickle=True)

            # Tracking the best model:
            if best_metric is None or AP.mean() > best_metric:
                best_metric = AP.mean()
                best_params = copy.deepcopy(model.state_dict())

        if epoch % opt.checkpoint_interval == 0:
            if best_params:
                torch.save(best_params, 'checkpoints/yolov3_ckpt_best.pth')
            else:
                torch.save(model.state_dict(), f"checkpoints/yolov3_ckpt_%d.pth" % epoch)



import tensorflow as tf


class Logger(object):
    def __init__(self, log_dir):
        """Create a summary writer logging to log_dir."""
        self.writer = tf.summary.create_file_writer(log_dir)

    def scalar_summary(self, tag, value, step):
        """Log a scalar variable."""
        with self.writer.as_default():
            tf.summary.scalar(tag, value, step=step)
            self.writer.flush()

    def list_of_scalars_summary(self, tag_value_pairs, step):
        """Log scalar variables."""
        with self.writer.as_default():
            for tag, value in tag_value_pairs:
                tf.summary.scalar(tag, value, step=step)
            self.writer.flush()



import torch
import numpy as np
import torch.nn.functional as F
import urllib
import cv2
import os


class SmileCNN(torch.nn.Module):
  """
  A simple convolutional network.
  
  Map from inputs with shape [batch_size, 1, height, width] to
  outputs with shape [batch_size, 1].
  """
  
  def __init__(self):
    super().__init__()

    self.conv1 = torch.nn.Conv2d(1, 128, kernel_size=7, padding=7//2, stride=2)
    self.conv2 = torch.nn.Conv2d(128, 64, kernel_size=5, padding=5//2, stride=2)
    self.conv3 = torch.nn.Conv2d(64, 32, kernel_size=5, padding=5//2)
    self.conv4 = torch.nn.Conv2d(32, 16, kernel_size=3, padding=3//2)
    
    self.fully_connected1 = torch.nn.Conv2d(16, 8, kernel_size=1)
    self.fully_connected2 = torch.nn.Conv2d(8, 2, kernel_size=1)

  def forward(self, x):
    max_pool = torch.nn.MaxPool2d(2, 2)

    x = max_pool(F.relu(self.conv1(x)))
    x = max_pool(F.relu(self.conv2(x)))
    x = max_pool(F.relu(self.conv3(x)))
    x = max_pool(F.relu(self.conv4(x)))

    x = self.fully_connected1(x)
    x = F.relu(x)
    x = self.fully_connected2(x)
    x = x.squeeze(3)
    x = x.squeeze(2)
    return x

# takes in a grayscale image
def squarify(image):
    rows, cols = image.shape
    if rows > cols:
        # max y > max x; make cols match rows
        diff = rows - cols
        edge1 = diff // 2
        # if split is slightly uneven, adjust
        edge2 = edge1 + diff % 2
        return np.concatenate((np.zeros((rows, edge1), dtype='float'), image, np.zeros((rows, edge2), dtype='float')), axis=1)
        
    elif rows < cols:
        # max x > max y; make rows match cols
        diff = cols - rows
        edge1 = diff // 2
        # if split is slightly uneven, adjust
        edge2 = edge1 + diff % 2
        return np.concatenate((np.zeros((edge1, cols), dtype='float'), image, np.zeros((edge2, cols), dtype='float')))

    # if haven't returned yet, image is already square
    return image
    

OUTPUT_SIZE = 64

def standardize(image):
    mean = image.mean(keepdims=True)
    stdev = image.std(keepdims=True)
    
    image = squarify((image - mean) / stdev)
    # it's a square, so rows = cols
    rows, _ = image.shape
    scale = OUTPUT_SIZE / rows
    warp = np.array([
        [scale, 0, 0],
        [0, scale, 0]
    ])
    return cv2.warpAffine(image, warp, dsize=(OUTPUT_SIZE, OUTPUT_SIZE),
        borderMode=cv2.BORDER_CONSTANT, borderValue=0)


# input: an image of a single face
def is_smiling(image, folder=None):
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    test_im = torch.from_numpy(np.float32(standardize(image)))[None, None, :, :]
    try:
        test_im = test_im.to(device)
        outputs = is_smiling.model(test_im)
    except AttributeError:
        # this is the first time we're running this, so initialize the neural network
        is_smiling.model = SmileCNN().to(device)
        is_smiling.model.float()

        if not os.path.isdir(folder):
            print('Creating new folder to download params into')
            os.makedirs(folder)

        param_path = os.path.join(folder, 'final_smile_net.pth')
        if not os.path.isfile(param_path) or not os.access(param_path, os.R_OK):
            print('Downloading parameters to path:', os.path.abspath(param_path))
            url = 'https://storage.googleapis.com/cv_final/model_dir/final_smile_net.pth'
            urllib.request.urlretrieve(url, param_path)
            print('Finished')

        is_smiling.model.load_state_dict(torch.load(param_path))
        outputs = is_smiling.model(test_im)
    
    soft = torch.nn.Softmax(1)
    return torch.max(outputs, 1)[1].item()



from __future__ import division

from face_detector.models import *
from face_detector.utils.utils import *
from face_detector.utils.datasets import *

import os
import sys
import time
import datetime
import argparse

from PIL import Image

import torch
from torch.utils.data import DataLoader
from torch.autograd import Variable

import matplotlib.pyplot as plt
import matplotlib.patches as patches
from matplotlib.ticker import NullLocator

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--image_folder", type=str, default="data/samples", help="path to dataset")
    parser.add_argument("--model_def", type=str, default="config/yolov3.cfg", help="path to model definition file")
    parser.add_argument("--weights_path", type=str, default="weights/yolov3.weights", help="path to weights file")
    parser.add_argument("--class_path", type=str, default="data/coco.names", help="path to class label file")
    parser.add_argument("--conf_thres", type=float, default=0.8, help="object confidence threshold")
    parser.add_argument("--nms_thres", type=float, default=0.4, help="iou thresshold for non-maximum suppression")
    parser.add_argument("--batch_size", type=int, default=1, help="size of the batches")
    parser.add_argument("--n_cpu", type=int, default=0, help="number of cpu threads to use during batch generation")
    parser.add_argument("--img_size", type=int, default=416, help="size of each image dimension")
    parser.add_argument("--checkpoint_model", type=str, help="path to checkpoint model")
    parser.add_argument("--add_labels", action='store_true', help="Put labels on bounding boxes")
    opt = parser.parse_args()
    print(opt)

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    os.makedirs("output", exist_ok=True)

    # Set up model
    model = Darknet(opt.model_def, img_size=opt.img_size).to(device)

    if opt.weights_path.endswith(".weights"):
        # Load darknet weights
        model.load_darknet_weights(opt.weights_path)
    else:
        # Load checkpoint weights
        model.load_state_dict(torch.load(opt.weights_path))

    model.eval()  # Set in evaluation mode

    dataloader = DataLoader(
        ImageFolder(opt.image_folder, img_size=opt.img_size),
        batch_size=opt.batch_size,
        shuffle=False,
        num_workers=opt.n_cpu,
    )

    classes = load_classes(opt.class_path)  # Extracts class labels from file

    Tensor = torch.cuda.FloatTensor if torch.cuda.is_available() else torch.FloatTensor

    imgs = []  # Stores image paths
    img_detections = []  # Stores detections for each image index

    print("\nPerforming object detection:")
    prev_time = time.time()
    for batch_i, (img_paths, input_imgs) in enumerate(dataloader):
        # Configure input
        input_imgs = Variable(input_imgs.type(Tensor)[:, :3, :, :])

        # Get detections
        with torch.no_grad():
            detections = model(input_imgs)
            detections = non_max_suppression(detections, opt.conf_thres, opt.nms_thres)

        # Log progress
        current_time = time.time()
        inference_time = datetime.timedelta(seconds=current_time - prev_time)
        prev_time = current_time
        print("\t+ Batch %d, Inference Time: %s" % (batch_i, inference_time))

        # Save image and detections
        imgs.extend(img_paths)
        img_detections.extend(detections)

    # Bounding-box colors
    cmap = plt.get_cmap("tab20b")

    print("\nSaving images:")
    # Iterate through images and save plot of detections
    for img_i, (path, detections) in enumerate(zip(imgs, img_detections)):

        print("(%d) Image: '%s'" % (img_i, path))

        # Create plot
        img = np.array(Image.open(path))
        plt.figure()
        fig, ax = plt.subplots(1)
        ax.imshow(img)

        # Draw bounding boxes and labels of detections
        if detections is not None:
            # Rescale boxes to original image
            detections = rescale_boxes(detections, opt.img_size, img.shape[:2])
            unique_labels = detections[:, -1].cpu().unique()
            n_cls_preds = len(unique_labels)
            for x1, y1, x2, y2, conf, cls_conf, cls_pred in detections:
                print("\t+ Label: %s, Conf: %.5f" % (classes[int(cls_pred)], cls_conf.item()))

                box_w = x2 - x1
                box_h = y2 - y1

                color = (0.0, 1.0, 0.0)
                # Create a Rectangle patch
                bbox = patches.Rectangle((x1, y1), box_w, box_h, linewidth=2, edgecolor=color, facecolor="none")
                # Add the bbox to the plot
                ax.add_patch(bbox)
                if opt.add_labels:
                    # Add label
                    plt.text(
                        x1,
                        y1,
                        s=classes[int(cls_pred)],
                        color="white",
                        verticalalignment="top",
                        bbox={"color": color, "pad": 0},
                    )

        # Save generated image with detections
        plt.axis("off")
        plt.gca().xaxis.set_major_locator(NullLocator())
        plt.gca().yaxis.set_major_locator(NullLocator())
        filename = path.split("/")[-1].split(".")[0]
        plt.savefig(f"output/{filename}.png", bbox_inches="tight", pad_inches=0.0)
        plt.close()



from eyeClassifierModel import EyeClassifierModel
import numpy as np
import torch
from torch import nn
from torch import optim
from torchvision import datasets, transforms, models
from torch.utils.data import Dataset, DataLoader
from torch.utils.data.sampler import SubsetRandomSampler

def load_test_data(datadir, trans, test_size = .2, batch=64):
    data = datasets.ImageFolder(datadir,transform=trans)
    dataloader = torch.utils.data.DataLoader(data, batch_size=batch, num_workers=4)
    return dataloader

# Data Normalization
def data_normalizer():
    return transforms.Compose([transforms.Resize(255),
        transforms.CenterCrop(224),
        transforms.ToTensor(),
        transforms.Normalize((0.485, 0.456, 0.406), (0.229, 0.224, 0.225)),])

if __name__ == '__main__':
    BATCH_SIZE=64
    data_dir = "./data/test_data/"

    trans = data_normalizer()
    dataloader = load_test_data(data_dir, trans, BATCH_SIZE)

    model = EyeClassifierModel()
    model.load_state_dict(torch.load('eye_classifier.pth'))
    model.eval()

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    model.to(device)

    criterion = nn.NLLLoss()
    optimizer = optim.Adam(model.parameters(), lr=0.001)

    loss = 0
    corrects = 0.0

    i = 0

    print(len(dataloader.dataset))

    for inputs, labels in dataloader:
        inputs, labels = inputs.to(device), labels.to(device)
        outputs = model(inputs)
        loss = criterion(outputs, labels)

        _, preds = torch.max(outputs, 1)

        loss += loss.item() * inputs.size(0)
        corrects += torch.sum(preds == labels.data)
        print(corrects)

        i += 1

    final_loss = loss / len(dataloader.dataset)
    final_acc = corrects.double() / len(dataloader.dataset)
    
    print('Loss: {:.4f} Acc: {:.4f}'.format(final_loss, final_acc))


from __future__ import division

import cv2
import os
import urllib
import matplotlib.pyplot as plt

from face_detector.models import *
from face_detector.utils.utils import *
from face_detector.utils.datasets import *
from torch.autograd import Variable


class FaceDetectorInterface:
    # Class is necessary to prevent having to load the model everytime run method is called
    def __init__(self, configs):
        self.cfigs = configs
        self.model = self.gen_model()
        self.classes = load_classes(self.cfigs['class_path'])

    def gen_model(self):
        model = Darknet(self.cfigs['model_arch'], img_size=self.cfigs['img_size'])
        model = model.to(self.cfigs['device'])
        model.load_state_dict(torch.load(self.cfigs['param_path'], map_location=self.cfigs['device']))
        model.eval()
        return model

    def run(self, image=None, image_path=None):
        '''
        Return a list of maps. Each map represents a face and contains information
        about it (corner coordinates and more).

        :param image: A ndarray with dimension  H x W x C
        :return: List of maps
        '''

        if image is None and image_path is None:
            raise ValueError
        if image_path:
            image = FaceDetectorInterface.read_image(image_path)


        Tensor = torch.cuda.FloatTensor if torch.cuda.is_available() else torch.FloatTensor


        H, W, C = image.shape
        img = transforms.ToTensor()(image)
        img, _ = pad_to_square(img, 0)
        img = resize(img, self.cfigs['img_size'])
        img = Variable(img.type(Tensor)[:3, :, :])
        img = img.unsqueeze(0)

        with torch.no_grad():
            batch_detections = self.model(img)
            batch_detections = non_max_suppression(batch_detections, self.cfigs['conf_thres'], self.cfigs['nms_thres'])

            detections = batch_detections[0]
            detections = rescale_boxes(detections, self.cfigs['img_size'], (H, W))

            results = []
            for detection in detections:
                x1, y1, x2, y2, conf, cls_conf, cls_pred = detection
                label = self.classes[int(cls_pred)]
                result = {
                    'x1': int(round(x1.item())),
                    'y1': int(round(y1.item())),
                    'x2': int(round(x2.item())),
                    'y2': int(round(y2.item())),
                    'label': label,
                    'object_confidence': conf.item(),
                    'class_confidence': cls_conf.item(),
                    'class_prediction': cls_pred.item()
                }
                results.append(result)
            return results

    @staticmethod
    def read_image(img_path):
        '''
        Returns an ndarray of dimensions H x W x C
        :param img_path: Path for image
        :return: Returns image including alpha channel
        '''

        return np.array(Image.open(img_path))

    @staticmethod
    def plot_faces(image, results, smile_classifier=None):
        for result in results:
            x1 = result['x1']
            x2 = result['x2']
            y1 = result['y1']
            y2 = result['y2']
            start = x1, y1
            end = x2, y2
            face = cv2.getRectSubPix(image, (x2 - x1, y2 - y1),
            ((x2 + x1) / 2, (y2 + y1) / 2))
            
            face = cv2.cvtColor(face, cv2.COLOR_RGB2GRAY)

            color = (0, 255, 0) if smile_classifier and smile_classifier(face) else (255, 0, 0)
            cv2.rectangle(image, start, end, color, 2)

        plt.figure()
        plt.imshow(image)
        plt.show()

    @staticmethod
    def create_old(param_path='face_detector/checkpoints/yolov3_ckpt_13.pth',
                   model_arch='face_detector/config/yolov3-custom.cfg',
                   class_path='face_detector/config/classes.names'):

        configs = FaceDetectorInterface.get_configs(param_path, model_arch, class_path)
        return FaceDetectorInterface(configs)

    @staticmethod
    def create(folder):

        if not os.path.isdir(folder):
            print('Creating new folder to download params into')
            os.makedirs(folder)

        param_path = os.path.join(folder, 'yolov3_ckpt_best.pth')
        model_arch = os.path.join(folder, 'yolov3-custom.cfg')
        class_path = os.path.join(folder, 'classes.names')

        if not os.path.isfile(param_path) or not os.access(param_path, os.R_OK):
            print('Downloading parameters to path:', os.path.abspath(param_path))
            url = 'https://storage.googleapis.com/cv_final/model_dir/yolov3_ckpt_best.pth'
            urllib.request.urlretrieve(url, param_path)
            print('Finished')
        if not os.path.isfile(model_arch) or not os.access(model_arch, os.R_OK):
            print('Downloading model architecture to path:', os.path.abspath(model_arch))
            url = 'https://storage.googleapis.com/cv_final/model_dir/yolov3-custom.cfg'
            urllib.request.urlretrieve(url, model_arch)
            print('Finished')
        if not os.path.isfile(class_path) or not os.access(class_path, os.R_OK):
            print('Downloading class names to path:', os.path.abspath(class_path))
            url = 'https://storage.googleapis.com/cv_final/model_dir/classes.names'
            urllib.request.urlretrieve(url, class_path)
            print('Finished')

        configs = FaceDetectorInterface.get_configs(param_path, model_arch, class_path)
        return FaceDetectorInterface(configs)

    @staticmethod
    def get_configs(param_path, model_arch, class_path):
        return {
            'param_path': param_path,
            'model_arch': model_arch,
            'class_path': class_path,
            'conf_thres': 0.8,
            'nms_thres': 0.4,
            'batch_size': 1,
            'n_cpu': 0,
            'img_size': 416,
            'device': torch.device("cuda" if torch.cuda.is_available() else "cpu")
        }


if __name__=='__main__':
    img_path = 'input_images/collage.png'
    interface = FaceDetectorInterface.create(folder='checkpoints')
    res = interface.run(image_path=img_path)
    print(res)



#!/usr/bin/env python3
from interface import EyeDetectorInterface
import cv2

if __name__ == "__main__":
    img = cv2.imread('../face_detector/input_images/small_img.png', cv2.IMREAD_GRAYSCALE)

    try:
        interface = EyeDetectorInterface()
        eye1, eye2, length = interface.detect(img) # img is the face image FaceDetector outputs
        print(eye1, eye2, length)
    except:
        print("Eyes not detected!")

    # my outputs for eye classification
    eye1_img = img[(eye1[0]-length):(eye1[0]+length), (eye1[1]-length):(eye1[1]+length)]
    eye2_img = img[(eye2[0]-length):(eye2[0]+length), (eye2[1]-length):(eye2[1]+length)]

    


    



from eye_classifier.eyeClassifierModel import EyeClassifierModel
import numpy as np
import torch
from torchvision import transforms
from torch.autograd import Variable
from PIL import Image
import urllib.request
import os


class EyeClassifierInterface:
    def __init__(self, param_path):
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.model = self.create_model(param_path)
        self.classes = ['close', 'open']

    def create_model(self, param_path):
        model = EyeClassifierModel()
        model.load_state_dict(torch.load(param_path,  map_location=self.device))
        model = model.to(self.device)
        model.eval()
        return model

    def read_image(self, img_path):
        image = Image.open(img_path)
        if image.mode != "RGB":
            image = image.convert("RGB")
        return image
    
    @staticmethod
    def create(folder):
        param_path = os.path.join(folder, 'eye_classifier.pth')

        if not os.path.isdir(folder):
            print('Creating new folder to download params into')
            os.makedirs(folder)

        if not os.path.isfile(param_path) or not os.access(param_path, os.R_OK):
            print('Downloading parameters to path:', os.path.abspath(param_path))
            url = 'https://www.dropbox.com/s/phlrvrvjwe41ycc/eye_classifier_2.pth?dl=1'
            urllib.request.urlretrieve(url, param_path)
            print('Finished')

        return EyeClassifierInterface(param_path)

    
    # Take image input
    # Return a dictionary, which contains: 
    #   classes: ['close', 'open']
    #   label corresponding to classification result (0 for close and 1 for open)
    #   probability of belonging to each class
    def run(self, image):
        image_transforms = transforms.Compose([transforms.Resize(255),
            transforms.CenterCrop(224),
            transforms.ToTensor(),
            transforms.Normalize((0.485, 0.456, 0.406), (0.229, 0.224, 0.225)),])
        image_tensor = image_transforms(image).float()
        image_tensor = image_tensor.unsqueeze_(0)
        input = Variable(image_tensor)
        input = input.to(self.device)
        output = self.model(input)
        output = torch.exp(output)
        index = output.data.cpu().numpy().argmax()

        result = {
            "classes": self.classes,
            "label": index,
            "probability": output.data.cpu().numpy()[0]
        }

        return result

if __name__=='__main__':
    img_path = 'open.jpg'
    interface = EyeClassifierInterface.create(folder='model_param')
    image = interface.read_image(img_path)
    res = interface.run(image)
    print(res)



import matplotlib.pyplot as plt
import numpy as np
import os


def plot(lis_path, y_label, x_label='Iterations'):
    data = np.load(lis_path, allow_pickle=True)
    x = np.arange(0, len(data))
    plt.figure()
    plt.scatter(x, data)
    plt.title('%s vs %s' % (x_label, y_label))
    plt.xlabel(x_label)
    plt.ylabel(y_label)
    plt.show()


def plot_all():
    root = './WIDER/data'
    data = [('train_loss_list.npy', 'Training Loss', 'Iterations'),
            ('val_f1.npy', 'F1 Score', 'Epochs'),
            ('val_mAP.npy', 'mAP Score', 'Epochs'),
            ('val_precision_list.npy', 'Precision', 'Epochs'),
            ('val_recall_list.npy', 'Recall', 'Epochs')]
    for file_name, y_label, x_label in data:
        path = os.path.join(root, file_name)
        plot(path, y_label, x_label)


def combine_data(data_1_pth, data_2_pth, out_path):
    data1 = np.load(data_1_pth, allow_pickle=True)
    data2 = np.load(data_2_pth, allow_pickle=True)
    data = np.concatenate((data1, data2))
    np.save(out_path, data, allow_pickle=True)


def combine_all():
    root = '/home/shuhao/Downloads/logs'
    data = ['train_loss_list', 'val_f1', 'val_mAP', 'val_precision_list', 'val_recall_list']
    for file_name in data:
        path1 = os.path.join(root, file_name + '_1.npy')
        path2 = os.path.join(root, file_name + '_2.npy')
        out_path = os.path.join(root, file_name + '.npy')
        combine_data(path1, path2, out_path)


if __name__=='__main__':
    # combine_all()
    plot_all()



from interface import EyeClassifierInterface


if __name__ == "__main__":
    img_path = 'open.jpg'
    
    folder = './'
    eye_classifier = EyeClassifierInterface.create(folder=folder)
    image = eye_classifier.read_image(img_path)
    res = eye_classifier.run(image)
    print(res)



#!/usr/bin/env python3

import os
import cv2
import numpy as np
import matplotlib.pyplot as plt
import urllib.request


class EyeDetectorInterface:

    def __init__(self, param_path):
        self.param_path = param_path
        self.model = cv2.ml.SVM_load(param_path)

    @staticmethod
    def create(folder):

        if not os.path.isdir(folder):
            print('Creating new folder to download params into')
            os.makedirs(folder)

        param_path = os.path.join(folder, 'svm_data.xml')

        if not os.path.isfile(param_path) or not os.access(param_path, os.R_OK):
            print('Downloading parameters to path:', os.path.abspath(param_path))
            url = 'https://storage.googleapis.com/cv_final/model_dir/svm_data.xml'
            urllib.request.urlretrieve(url, param_path)
            print('Finished')
            pass

        return EyeDetectorInterface(param_path)

    # Stores hog value of img into data[data_idx, :] and label into labels[data_idx, 0]
    def get_hog(self, img, data_idx, data, label, labels):
        winSize = (40,40)
        blockSize = (16,16)
        blockStride = (8,8)
        cellSize = (8,8)
        nbins = 16
        derivAperture = 1
        winSigma = 4.
        histogramNormType = 0
        L2HysThreshold = 2.0000000000000001e-01
        gammaCorrection = 0
        nlevels = 64

        hog = cv2.HOGDescriptor(winSize,blockSize,blockStride,cellSize,nbins,derivAperture,winSigma,
                                histogramNormType,L2HysThreshold,gammaCorrection,nlevels)
        hog_val = hog.compute(img)
        data[data_idx, :] = hog_val.reshape(1024,)
        labels[data_idx, 0] = label
        data_idx += 1
        return data_idx

    def check_border(self, coor, img0_x, img0_y, dist):
        if (coor[0] < dist or coor[1] < dist):
            return True
        elif (coor[1] > (img0_x - dist) or coor[0] > (img0_y - dist)):
            return True
        return False

    def get_range(self, coor, dist):
        x_low = coor[1] - dist
        x_high = coor[1] + dist
        y_low = coor[0] - dist
        y_high = coor[0] + dist
        return x_low, x_high, y_low, y_high

    def preprocess_bioid(self, bioid_path, patch_diameter):
        img0 = cv2.imread(bioid_path + 'BioID_0020.pgm', cv2.IMREAD_GRAYSCALE)
        pos_data0 = open(bioid_path + "BioID_0020.eye", "r")
        lx, ly, rx, ry = (pos_data0.read()[13:]).split('\t')

        # Pre-process all BioID Face data
        pos_data = [] # stores all positive image patches
        neg_data = [] # stores all negative image patches

        neg_offset = patch_diameter
        dist = patch_diameter // 2
        img0_x, img0_y = img0.shape
        x = np.arange(img0_x)
        y = np.arange(img0_y)
        xx, yy = np.meshgrid(x, y)
        coors = np.concatenate([xx.reshape(-1,1), yy.reshape(-1, 1)], 1)

        data = np.zeros([15200, 1024]) # now oversized
        data_idx = 0
        labels = np.zeros([15200, 1])

        for i in range(1520):
            sample_pgm = bioid_path + 'BioID_' + str(i).zfill(4) + '.pgm'
            sample_eye = bioid_path + 'BioID_' + str(i).zfill(4) + '.eye'
            img = cv2.imread(sample_pgm, cv2.IMREAD_GRAYSCALE)
            
            pos_sample = open(sample_eye, "r")
            lx, ly, rx, ry = (pos_sample.read()[13:]).split('\t')
            left_eye = np.array([int(lx), int(ly)])
            right_eye = np.array([int(rx), int(ry)])
            
            # append left eye
            if not self.check_border(left_eye, img0_x, img0_y, dist):
                x_low, x_high, y_low, y_high = self.get_range(left_eye, dist)
                patch = cv2.resize(img[x_low:x_high, y_low:y_high], (40, 40))
                pos_data.append(patch)
                data_idx = self.get_hog(pos_data[-1], data_idx, data, 1, labels)
            # append right eye
            if not self.check_border(right_eye, img0_x, img0_y, dist):
                x_low, x_high, y_low, y_high = self.get_range(right_eye, dist)
                patch = cv2.resize(img[x_low:x_high, y_low:y_high], (40, 40))
                pos_data.append(patch)
                data_idx = self.get_hog(pos_data[-1], data_idx, data, 1, labels)
            
            # sample first, then check if within eye regions
            neg_idx = np.random.randint(0,coors.shape[0],size=3)
            for idx in neg_idx:
                neg_sample = coors[idx]
                # check if within eye regions
                if (left_eye[0] - neg_offset < neg_sample[0] < right_eye[0] + neg_offset) and (left_eye[1] - neg_offset < neg_sample[1] < right_eye[1] + neg_offset):
                    continue
                # check if within border regions
                elif not self.check_border(neg_sample, img0_x, img0_y, dist):
                    x_low, x_high, y_low, y_high = self.get_range(neg_sample, dist)
                    patch = cv2.resize(img[x_low:x_high, y_low:y_high], (40, 40))
                    neg_data.append(patch)
                    data_idx = self.get_hog(neg_data[-1], data_idx, data, -1, labels)
        return pos_data, neg_data, data[:data_idx,:], labels[:data_idx,:]

    # Helper to split data into 9 training folds and 1 testing fold
    def split_bioid(self, pos_data, neg_data, data, labels):
        # Split data for training and testing
        total_data = len(pos_data) + len(neg_data)
        test_total = total_data // 10 # 1/10 for testing

        # Create dictionary to store 10 different splits
        data_dict = {'train_data':[],'test_data':[], 'train_labels':[], 'test_labels':[]}
        for i in range(10):
            data_dict['test_data'].append(data[(i * test_total):((i + 1) * test_total), :])
            data_dict['test_labels'].append(labels[(i * test_total):((i + 1) * test_total), :])
            train_data0 = data[0:(i * test_total), :].reshape(-1, 1024)
            train_data1 = data[((i + 1) * test_total):total_data, :].reshape(-1, 1024)
            train_labels0 = labels[0:(i * test_total), :].reshape(-1, 1)
            train_labels1 = labels[((i + 1) * test_total):total_data, :].reshape(-1, 1)
            data_dict['train_data'].append(np.concatenate((train_data0, train_data1), axis = 0))
            data_dict['train_labels'].append(np.concatenate((train_labels0, train_labels1), axis = 0))
        return data_dict

    # bioid path is where the BioID images are stored
    def train(self, bioid_path):
        if (bioid_path[-1] != '/'):
            bioid_path = bioid_path + '/'
        _, _, data, labels = self.preprocess_bioid(bioid_path)

        # Set training parameters
        svm = cv2.ml.SVM_create()
        svm.setKernel(cv2.ml.SVM_LINEAR)
        svm.setType(cv2.ml.SVM_C_SVC)
        svm.setC(0.7) # soft margin previously 2.67
        svm.setGamma(6) # radial basis function previously 5.383

        svm.train(data.astype(np.float32), cv2.ml.ROW_SAMPLE, labels.astype(np.int32))
        svm.save('svm_data.xml')

    # Test overall detection model
    def test(self, bioid_path):
        svm = cv2.ml.SVM_create()
        svm.setKernel(cv2.ml.SVM_LINEAR)
        svm.setType(cv2.ml.SVM_C_SVC)
        svm.setC(0.7) # soft margin previously 2.67
        svm.setGamma(6) # radial basis function previously 5.383
        
        pos_data, neg_data, data, labels = self.preprocess_bioid(bioid_path)
        total_data = len(pos_data) + len(neg_data)
        test_total = total_data // 10 # 1/10 for testing
        data_dict = self.split_bioid(pos_data, neg_data, data, labels)

        accuracies = []
        for i in range(10):
            train_data = data_dict['train_data'][i]
            test_data = data_dict['test_data'][i]
            train_labels = data_dict['train_labels'][i]
            test_labels = data_dict['test_labels'][i]
            svm.train(train_data.astype(np.float32), cv2.ml.ROW_SAMPLE, train_labels.astype(np.int32))
            svm.save('svm_data.xml')
            results = np.zeros([test_total, 1])
            for i in range(test_total):
                results[i, 0] = svm.predict(test_data[i,:].reshape(1,-1).astype(np.float32))[1]

            mask = results==test_labels
            correct = np.count_nonzero(mask)
            accuracies.append(correct / test_total)
            print(accuracies[-1])

    # Cross validation on binary classification model (SVM)
    def train_and_test(self, bioid_path, scales=[40]):
        # Set training parameters
        svm = cv2.ml.SVM_create()
        svm.setKernel(cv2.ml.SVM_LINEAR)
        svm.setType(cv2.ml.SVM_C_SVC)
        svm.setC(0.7) # soft margin previously 2.67
        svm.setGamma(6) # radial basis function previously 5.383
        
        if (len(scales) == 0):
            print("Must choose at least one scale!! If you don't know where to start, leave this parameter empty.")
            return

        pos_data, neg_data, data, labels = self.preprocess_bioid(bioid_path, scales[0])
        if (len(scales) > 1):
            for scale in scales[1:]:
                new_pos_data, new_neg_data, new_data, new_labels = self.preprocess_bioid(bioid_path, scale)
                pos_data = pos_data + new_pos_data
                neg_data = neg_data + new_neg_data
                data = np.vstack((data, new_data))
                labels = np.vstack((labels, new_labels))

        total_data = len(pos_data) + len(neg_data)
        test_total = total_data // 10 # 1/10 for testing
        data_dict = self.split_bioid(pos_data, neg_data, data, labels)

        accuracies = []
        for i in range(10):
            train_data = data_dict['train_data'][i]
            test_data = data_dict['test_data'][i]
            train_labels = data_dict['train_labels'][i]
            test_labels = data_dict['test_labels'][i]
            svm.train(train_data.astype(np.float32), cv2.ml.ROW_SAMPLE, train_labels.astype(np.int32))
            svm.save('svm_data.xml')
            results = np.zeros([test_total, 1])
            for i in range(test_total):
                results[i, 0] = svm.predict(test_data[i,:].reshape(1,-1).astype(np.float32))[1]

            mask = results==test_labels
            correct = np.count_nonzero(mask)
            accuracies.append(correct / test_total)
            print(accuracies[-1])
   

   # Sequential Labeling Algorithm to segment the coordinates
    def seq_labeling(self, y_coords, x_coords, stride):
        labeled = {} # coor: label
        equivalence = [] # a list of equivalence sets
        offsets = np.array([[-stride,-stride],[0,-stride],[stride,-stride],
                            [-stride,0],[stride,0],[-stride,stride],
                            [0,stride],[stride,stride]])
        labeled_flag = 0
        num_labels = 0
        for i in range(y_coords.size):
            labeled_flag = 0
            xx = x_coords[i]
            yy = y_coords[i]
            neighbors = (np.array([xx,yy]) + offsets).tolist()
            for n in neighbors:
                nn = (n[0],n[1])
                if labeled_flag == 0:
                    if nn in labeled:
                        labeled[(xx,yy)] = labeled[nn]
                        labeled_flag = 1
                elif nn in labeled: # (xx,yy) already labeled
                    if labeled[nn] != labeled[(xx,yy)]:
                        noted = 0 # flag
                        # note in our list of equivalence sets
                        for e in equivalence:
                            if labeled[nn] in e:
                                e.add(labeled[(xx,yy)])
                                noted = 1
                            elif labeled[(xx,yy)] in e:
                                e.add(labeled[nn])
                                noted = 1
                        if noted == 0:
                            equivalence.append({labeled[nn], labeled[(xx,yy)]})
            
            # if no neighbor has been labeled
            if labeled_flag == 0:
                labeled[(xx,yy)] = num_labels # adds new label
                num_labels+= 1
        
        # one final pass to adjust for equivalencies
        for l in labeled:
            cur_label = labeled[l]
            for e in equivalence:
                if cur_label in e:
                    labeled[l] = np.min(list(e))
        return labeled

    # Finds 2 labels that correspond to the 2 eyes 
    def find_eyes(self, labeled, m):
        labels = {}
        for l in labeled:
            if labeled[l] not in labels:
                labels[labeled[l]] = {l}
            else:
                labels[labeled[l]].add(l)
        
        ind_scores = {} # evaluate each label individually
        centers = {}
        for l in labels:
            y_average = np.sum(np.array(list(labels[l]))[:,1]) / np.array(list(labels[l])).shape[0]
            x_average = np.sum(np.array(list(labels[l]))[:,0]) / np.array(list(labels[l])).shape[0]
            centers[l] = np.array([x_average, y_average]).astype(np.int)
            # 1. maximizes number of votes (prioritized)
            # 2. minimizes average y value
            ind_scores[l] = len(labels[l]) / (y_average**2)
        
        if (len(labels) < 3):
            return list(centers.values())
        
        # Evaluate all pairs and encourage
        # 1. minimize difference between y values
        # 2. maximize difference between x values
        # 3. maximize the 2's individual scores
        scores = {}
        for l1 in labels:
            for l2 in labels:
                if (l1 == l2):
                    continue
                y_diff = np.abs(centers[l1][1] - centers[l2][1])
                x_diff = np.abs(centers[l1][0] - centers[l2][0])
                if (y_diff == 0):
                    y_diff = 0.01
                scores[(l1, l2)] = ind_scores[l1] * ind_scores[l2] * (x_diff**2) / y_diff
        
        res_labels = sorted(scores, key=scores.get, reverse=True)[0]
        return [centers[res_labels[0]], centers[res_labels[1]]]

    # param_path is the path is where svm_data.xml is stored
    def detect(self, img):
        svm = self.model
        winSize = (40,40)
        blockSize = (16,16)
        blockStride = (8,8)
        cellSize = (8,8)
        nbins = 16
        derivAperture = 1
        winSigma = 4.
        histogramNormType = 0
        L2HysThreshold = 2.0000000000000001e-01
        gammaCorrection = 0
        nlevels = 64

        hog = cv2.HOGDescriptor(winSize,blockSize,blockStride,cellSize,nbins,derivAperture,winSigma,
                                histogramNormType,L2HysThreshold,gammaCorrection,nlevels)
        m, n = img.shape
        
        # for images that are big enough
        step = max(1, n // 90)
        length = step * 20
        stride = 4 # to cut down on computation costs
        
        if (m < 20 or n < 20):
            return None

        if (n < 35):
            length = step * 5
            stride = 1
        elif (n < 50):
            length = step * 7
            stride = 1
        elif (n < 90):
            length = step * 12
            stride = 1
        elif (n < 120):
            length = step * 16
            stride = 2
        
        x_range = length + np.array([stride * n for n in range((m - 2 * length - 2) // stride)])
        y_range = length + np.array([stride * n for n in range((n - 2 * length - 2) // stride)])
        res = np.zeros([m, n])

        for x in x_range:
            for y in y_range:
                sub_img = cv2.resize(img[(x-length):(x+length):step, (y-length):(y+length):step], (40, 40))
                temp = svm.predict(hog.compute(sub_img).astype(np.float32).reshape(1,-1))[1][0][0]
                res[x,y] = 0 if temp == -1 else 1
                
        mask = res>0
        y_coords, x_coords = np.nonzero(mask)
        
        # Call Sequential Labeling function
        labeled = self.seq_labeling(y_coords, x_coords, stride)
        
        # Call method to locate the eyes
        try:
            eyes = self.find_eyes(labeled, m)
            if (len(eyes) == 0):
                return None
            return (eyes, length)
        except:
            return None



from setuptools import setup, find_packages


REQUIRED_PACKAGES = ['opencv-python>=4.4.0.46',
                     'torch>=1.6.0',
                     'numpy>=1.19.1',
                     'torchvision>=0.7.0',
                     'matplotlib>=3.3.1',
                     'Pillow>=8.0.1']

setup(
    name='LRSPhotos',
    version='0.1.3',
    author="Shuhao Lai",
    author_email="Shuhaolai18@gmail.com",
    description="Face and eye detector and eye and smile classifier bundled into one.",
    packages=find_packages(include=['face_detector', 'face_detector.*',
                                    'eye_classifier', 'eye_classifier.*',
                                    'eye_detector', 'eye_detector.*',
                                    'smile_classifier', 'smile_classifier.*',
                                    'lrsphotos', 'lrsphotos.*']),
    install_requires=REQUIRED_PACKAGES
)